#!/usr/bin/env python
#-*- coding:utf-8 -*-
"""
z2m.dll wrapper + some useful functions
Mifare 4K: 0-31 - 3x16 + trailer;  32-39 - 15x16 + trailer
"""

import sys
import time

from ctypes import *
from struct import pack, unpack

__all__ = ("server",)

reader = None
DLL = None

class DLLka:
    def __init__(self, path):
        self.dll = windll.LoadLibrary(path)

    def __getattr__(self, func, *v):
        for i in range(0, 33, 4):
            try:
                f = getattr(self.dll, "%s@%d" % (func, i))
            except:
                pass
            else:
                break
        return f

class Card:
    def __init__(self, card_id, info):
        self.items = info
        self.id = card_id

    def __getitem__(self, item):
        return self.items[item]

    def __contains__(self, item):
        return item in self.items

    def read(self, sector, block, key_type=False, key="\xff"*6):
        global reader

        z2mSetIndicatorState(reader, 1, 2, 2)
        sector_id = z2mOpenMfSector(self.id, sector, key_type, key)
        if sector_id <= 32:
            return None

        block = z2mReadMfBlock(sector_id, block)
        if hasattr(block, "raw"):
            data = block.raw
        else:
            data = None

        z2mCloseMfSector(sector_id)
        z2mSetIndicatorState(reader, 1, 1, 2)
        return data

    def write(self, sector, block, data, key_type=False, key="\xff"*6):
        global reader

        z2mSetIndicatorState(reader, 1, 2, 2)
        sector_id = z2mOpenMfSector(self.id, sector, key_type, key)
        if sector_id <= 32:
            return None

        block = z2mWriteMfBlock(sector_id, block, data)
        if hasattr(block, "raw"):
            data = block.raw
        else:
            data = None

        z2mCloseMfSector(sector_id)
        z2mSetIndicatorState(reader, 1, 1, 2)
        return data        

# -----------------------------------------------------------------------------
# FUNCS
# -----------------------------------------------------------------------------

def init():
    global DLL
    DLL = DLLka("z2m.dll")

def server(callback):
    global card_id, reader

    z2mInitialize()
    res = z2mFindSoleReader()
    if "sn" not in res:
        return "FAIL"

    reader = z2mOpenReader(res['port'], res['sn'], res['type'])

    lastcard = False
    types = ['Unknown!', 'Mifare UL', 'Mifare 1K', 'Mifare 4K']
    sectors = [None, None, 64, 256]
    while True:
        z2mSetIndicatorState(reader, 2, 2, 2)

        while True:
            time.sleep(0.05)
            res = z2mFindCard(reader)
            if lastcard != False and res['result'] == 5:
                lastcard = False
# wtf?      if res['result'] != 0 and res['result'] != 5:
#                print '!'
            if res['result'] == 0 and res['num'] != lastcard:
                break

        z2mSetIndicatorState(reader, 1, 1, 2)
        lastcard = res['num']
        
        card = {"type": types[res['type']],
                "ntype": res['type'],
                "id": res['num'].encode("hex"),
                "sectors": sectors[res['type']],
                "err": None}

        card_id = z2mOpenCard(reader, res['type'], res['num'])

        if card_id <= 32:
            continue

        if res['type'] == 1:
            z2mSetIndicatorState(reader, 1, 2, 2)
            data = z2mReadCardData(card_id, 0, 64)
            if type(data) == str:
                card['ul'] = data
                callback(Card(card_id, card))
            z2mSetIndicatorState(reader, 1, 1, 2)
        elif res['type'] == 2 or res['type'] == 3:
            callback(Card(card_id, card))
        z2mCloseCard(card_id)
    return "WTF"

# -----------------------------------------------------------------------------
# WRAPPERS
# -----------------------------------------------------------------------------

def z2mInitialize():
    return DLL.z2mInitialize()

def z2mFinalize():
    return DLL.z2mFinalize()

def z2mFindSoleReader():
    VHandle = create_string_buffer(4)
    szPort = create_string_buffer(16)
    VInfo = pack("<7I", addressof(szPort), 16, 0, 0, 0, 0, 0)
    result = DLL.z2mFindFirstReaderA(VHandle, 0, VInfo, 0)
    if result:
        return {'result': result}
    VHandle = unpack("<I", VHandle)[0]
    DLL.z2mCloseFindReader(VHandle)
    b, c, nTip, nSN = unpack("<4I", VInfo[:16])
    return {'result': result, 'port': szPort.value, 'type': nTip, 'sn': nSN}

def z2mOpenReader(port, sn, typ):
    VHandle = create_string_buffer(4)
    result = DLL.z2mOpenReaderA(VHandle, port, sn, typ, 0 , 0)
    if result:
        return result
    VHandle = unpack("<I", VHandle)[0]
    return VHandle

def z2mGetIndicatorState(AHandle):
    red, green, sound = [create_string_buffer(8) for i in range(3)]
    result = DLL.z2mGetIndicatorState(AHandle, red, green, sound, 0)
    if result:
        return {'result': result}
    aRed = aGreen = aSound = {}
    aRed['enable'] = not (red[0] == "\0"); aRed['auto'] = not (red[4] == "\0")
    aGreen['enable'] = not (green[0] == "\0"); aGreen['auto'] = not (green[4] == "\0")
    aSound['enable'] = not (sound[0] == "\0"); aSound['auto'] = not (sound[4] == "\0")
    return {'result': result, 'red': aRed, 'green': aGreen, 'sound': aSound}

def z2mSetIndicatorState(AHandle, red, green, sound):
    return DLL.z2mSetIndicatorState(AHandle, red, green, sound, 0)

def z2mAddIndicatorFlash(AHandle, flashArr, AReset):
    AFlashArr = "";
    AArrCount = len(flashArr)
    for flash in flashArr:
        AFlashArr += pack("<4I", flash['red'], flash['green'], flash['sound'], flash['time'])
    AReset = -1 if AReset else 0
    return DLL.z2mAddIndicatorFlash, (AHandle, AFlashArr, AArrCount, AReset, 0, 0)

def z2mFindCard(ARd):
    VInfo = create_string_buffer(20)
    result = DLL.z2mFindCard(ARd, VInfo, 0)
    if result:
        return {'result': result}
    VInfo = VInfo.raw
    num = VInfo[1:1+ord(VInfo[0])];
    typ = unpack("<I", VInfo[16:20])[0]
    return {'result': result, 'num': num, 'type': typ}

def z2mOpenCard(ARd, ACdType, ACdNum):
    VHandle = create_string_buffer(4)
    ACdNum = chr(len(ACdNum)) + ACdNum
    ACdNum = create_string_buffer(ACdNum.ljust(16, "\x00"))
    result = DLL.z2mOpenCard(VHandle, ARd, ACdType, ACdNum, 0)
    if result:
        return result
    VHandle = unpack("<I", VHandle[:4])[0]
    return VHandle

def z2mReadCardData(AHandle, AAddr, ACount):
    VBuf = create_string_buffer("\x00" * ACount)
    VRead = create_string_buffer(4)
    result = DLL.z2mReadCardData(AHandle, AAddr, VBuf, ACount, VRead, 0)
    if result:
        return result
    VRead = unpack("<I", VRead[:4])[0]
    return VBuf[:VRead]

def z2mCloseCard(AHandle):
    return DLL.z2mCloseCard(AHandle)

def z2mOpenMfSector(hCard, nSectN, useKeyB, nKey):
    VHandle = create_string_buffer(4)
    nKey = nKey.ljust(8, "\x00")
    AInfo = pack("<6I", hCard, nSectN, 0, 1 if useKeyB else 0, 0, 0) + nKey + "\x00"*16 + "\x00"*8
    result = DLL.z2mOpenMfSector(VHandle, AInfo, 0)
    if result:
        return result
    VHandle = unpack("<I", VHandle[:4])[0]
    return VHandle

def z2mReadMfBlock(AHandle, ABlockN):
    VBuf16 = create_string_buffer(16)
    result = DLL.z2mReadMfBlock(AHandle, ABlockN, VBuf16, 0)
    if result:
        return result
    return VBuf16

def z2mWriteMfBlock(AHandle, ABlockN, data):
    try:
        VBuf16 = create_string_buffer(data, 17)
        result = DLL.z2mWriteMfBlock(AHandle, ABlockN, VBuf16, 0)
    except:
        return None
    return result

def z2mCloseMfSector(AHandle):
    return DLL.z2mCloseMfSector(AHandle)

# -----------------------------------------------------------------------------
# EXAMPLE
# -----------------------------------------------------------------------------

init()

if __name__ == "__main__":
    print "Init lib... ", z2mInitialize()

    print "Find reader... ",
    sys.stdout.flush()
    res = z2mFindSoleReader()
    if "sn" not in res:
        print "FAIL (disconnected or busy)", res['result']
        quit()
        
    print res['result'], " ", res['sn'], " @ ", res['port']

    reader = z2mOpenReader(res['port'], res['sn'], res['type'])
    print "Open reader... ", hex(reader)

    lastcard = False
    types = ['Unknown!', 'Mifare UL', 'Mifare 1K', 'Mifare 4K']

    while True:
        print "\nWaiting for card... ",
        sys.stdout.flush()
        z2mSetIndicatorState(reader, 2, 2, 2)

        while True:
            time.sleep(0.05)
            res = z2mFindCard(reader)
            if lastcard != False and res['result'] == 5:
                lastcard = False
            if res['result'] != 0 and res['result'] != 5:
                print '!'
            if res['result'] == 0 and res['num'] != lastcard:
                break

        z2mSetIndicatorState(reader, 1, 1, 2)
        print types[res['type']], " ", res['num'].encode("hex")

        lastcard = res['num']

        print " Open card... ",
        sys.stdout.flush()
        card = z2mOpenCard(reader, res['type'], res['num'])
        print hex(card)

        if card > 32:
            if res['type'] == 1:
                z2mSetIndicatorState(reader, 1, 2, 2)
                print "  Read UL... ", repr(z2mReadCardData(card, 0, 64))
                z2mSetIndicatorState(reader, 1, 1, 2)
            elif res['type'] == 2 or res['type'] == 3:
                print "  Open sector... ", 
                sys.stdout.flush()
                sector = z2mOpenMfSector(card, 0, False, "\xff\xff\xff\xff\xff\xff")
                print sector
                if sector > 32:
                    z2mSetIndicatorState(reader, 1, 2, 2)
                    print "  Read block... ",
                    sys.stdout.flush()
                    block = z2mReadMfBlock(sector, 0)
                    if hasattr(block, "raw"):
                        print "OK:", repr(block.raw)
                    else:
                        print "FAIL"
                    print "  Close sector... ", z2mCloseMfSector(sector)
                    z2mSetIndicatorState(reader, 1, 1, 2)

            print " Close card... ", z2mCloseCard(card)
